@@ -1,3 +1,7 @@
+"""
+This module defines a base class for paths in a *cloud* storage, aka "blob store".
+This is in contrast to a *local* disk storage, which is the subject of `_local.py`.
+"""
 import pathlib
 from ._upath import Upath
 from ._local import LocalUpath
@@ -18,11 +22,15 @@ def _resolve_local_path(p):
 class BlobUpath(Upath, EnforceOverrides):
     @property
     def blob_name(self) -> str:
-        return self._path.lstrip('/')
+        return self._path.lstrip("/")
 
-    def download_dir(self, target, *, overwrite=False) -> int:
+    def download_dir(self, target, *, overwrite=False, desc=None) -> int:
         target_ = _resolve_local_path(target)
-        return self.export_dir(target_, overwrite=overwrite)
+        return self.export_dir(
+            target_,
+            overwrite=overwrite,
+            desc=desc or f"Downloading from {self!r} into {target_!r}",
+        )
 
     def download_file(self, target, *, overwrite=False) -> None:
         target_ = _resolve_local_path(target)
@@ -30,7 +38,7 @@ class BlobUpath(Upath, EnforceOverrides):
 
     @overrides
     def is_dir(self) -> bool:
-        '''In a typical blob store, there is no such concept as a
+        """In a typical blob store, there is no such concept as a
         "directory". Here we emulate the concept in a local file
         system. If there is a blob named like
 
@@ -47,7 +55,7 @@ class BlobUpath(Upath, EnforceOverrides):
         Consequently, `is_dir` is equivalent
         to "having stuff in the dir". There is no such thing as
         an "empty directory" in blob stores.
-        '''
+        """
         try:
             next(self.iterdir())
             return True
@@ -58,23 +66,25 @@ class BlobUpath(Upath, EnforceOverrides):
     def iterdir(self):
         # A naive, inefficient implementation.
         p0 = self._path  # this could be '/'.
-        if not p0.endswith('/'):
-            p0 += '/'
+        if not p0.endswith("/"):
+            p0 += "/"
         np0 = len(p0)
         subdirs = set()
         for p in self.riterdir():
             tail = p._path[np0:]
-            if tail.startswith('/'):
+            if tail.startswith("/"):
                 raise Exception(f"malformed blob name '{p._path}'")
-            if '/' in tail:
-                tail = tail[: tail.find('/')]
+            if "/" in tail:
+                tail = tail[: tail.find("/")]
             if tail not in subdirs:
                 yield self / tail
                 subdirs.add(tail)
 
-    def upload_dir(self, source, *, overwrite=False) -> int:
+    def upload_dir(self, source, *, overwrite=False, desc=None) -> int:
         s = _resolve_local_path(source)
-        return self.import_dir(s, overwrite=overwrite)
+        return self.import_dir(
+            s, overwrite=overwrite, desc=desc or f"Uploading from {s!r} into {self!r}"
+        )
 
     def upload_file(self, source, *, overwrite=False) -> None:
         s = _resolve_local_path(source)
