@@ -1,4 +1,5 @@
 from __future__ import annotations
+
 # Enable using `Upath` in type annotations in the code
 # that defines this class.
 # https://stackoverflow.com/a/49872353
@@ -12,14 +13,20 @@ from io import BufferedReader, UnsupportedOperation, BytesIO
 from typing import Union
 
 import opnieuw
+import requests
 from google import resumable_media
 from google.oauth2 import service_account
 from google.cloud import storage
 from google.api_core.exceptions import (
-    NotFound, PreconditionFailed, TooManyRequests)
+    NotFound,
+    PreconditionFailed,
+    TooManyRequests,
+    GatewayTimeout,
+    ServiceUnavailable,
+)
 from overrides import overrides
 
-from ._upath import FileInfo, Upath, LockAcquisitionTimeoutError
+from ._upath import FileInfo, Upath, LockAcquireError, LockReleaseError
 from ._blob import BlobUpath
 from ._local import LocalUpath
 
@@ -30,16 +37,28 @@ logger = logging.getLogger(__name__)
 # 67108864 = 256 * 1024 * 256 = 64 MB
 MEGABYTES32 = 33554432
 MEGABYTES64 = 67108864
+LARGE_FILE_SIZE = MEGABYTES64
+
+
+RETRY_WRITE_ON_EXCEPTIONS = (
+    TooManyRequests,
+    GatewayTimeout,
+    ServiceUnavailable,
+    requests.ReadTimeout,
+)
 
 
 class GcpBlobUpath(BlobUpath):
-    def __init__(self, *paths: str,
-                 bucket_name: str = None,
-                 project_id: str = None,
-                 credentials: service_account.Credentials = None,
-                 ):
-        '''
+    def __init__(
+        self,
+        *paths: str,
+        bucket_name: str = None,
+        project_id: str = None,
+        credentials: service_account.Credentials = None,
+    ):
+        """
         If you have GCP account_info in a dict with these elements:
+
             'type': 'service_account',
             'project_id':
             'private_key_id':
@@ -63,19 +82,19 @@ class GcpBlobUpath(BlobUpath):
 
         TODO: check out github repo `googleapis/python-cloud-core` to see whether there is
         a way or a need to infer these values. In particular, see class `google.cloud.client.ClientWithProject`.
-        '''
+        """
         if bucket_name is None:
             assert len(paths) == 1
             path = paths[0]
-            assert path.startswith('gs://')
+            assert path.startswith("gs://")
             path = path[5:]
-            k = path.find('/')
+            k = path.find("/")
             if k < 0:
                 bucket_name = path
-                paths = ('/', )
+                paths = ("/",)
             else:
                 bucket_name = path[:k]
-                paths = (path[k:], )
+                paths = (path[k:],)
 
         super().__init__(*paths)
         self.bucket_name = bucket_name
@@ -91,11 +110,11 @@ class GcpBlobUpath(BlobUpath):
         return "{}('gs://{}/{}')".format(
             self.__class__.__name__,
             self.bucket_name,
-            self._path,
+            self._path.lstrip("/"),
         )
 
     def __str__(self) -> str:
-        return self.__repr__()
+        return f"gs://{self.bucket_name}/{self._path.lstrip('/')}"
 
     def __eq__(self, other) -> bool:
         if other.__class__ is not self.__class__:
@@ -137,17 +156,17 @@ class GcpBlobUpath(BlobUpath):
         # (when not None) can't be pickled.
         # the `service_account.Credentials` class object can be pickled.
         return {
-            '_path': self._path,
-            'bucket_name': self.bucket_name,
-            '_project_id': self._project_id,
-            '_credentials': self._credentials,
+            "_path": self._path,
+            "bucket_name": self.bucket_name,
+            "_project_id": self._project_id,
+            "_credentials": self._credentials,
         }
 
     def __setstate__(self, data):
-        self._path = data['_path']
-        self.bucket_name = data['bucket_name']
-        self._project_id = data['_project_id']
-        self._credentials = data['_credentials']
+        self._path = data["_path"]
+        self.bucket_name = data["bucket_name"]
+        self._project_id = data["_project_id"]
+        self._credentials = data["_credentials"]
         self._client = None
         self._bucket = None
         self._blob = None
@@ -170,21 +189,21 @@ class GcpBlobUpath(BlobUpath):
         return self._bucket
 
     def blob(self):
-        '''
+        """
         This constructs a Blob object irrespecitive of whether the blob
         exists in cloud storage.
-        '''
+        """
         if self._blob is None:
             self._blob = self.bucket.blob(self.blob_name)
         return self._blob
 
     def get_blob(self):
-        '''
+        """
         While `.blob` simply constructs a Blob object,
         `.get_blob` makes network calls to refresh properties
         of the object in cloud storage. If the blob does not exist,
         return `None`.
-        '''
+        """
         b = self.blob()
         try:
             b.reload(client=self.client)
@@ -193,9 +212,9 @@ class GcpBlobUpath(BlobUpath):
             return None
 
     @opnieuw.retry(
-        retry_on_exceptions=(TooManyRequests, ),
+        retry_on_exceptions=RETRY_WRITE_ON_EXCEPTIONS,
         max_calls_total=10,
-        retry_window_after_first_call_in_seconds=60,
+        retry_window_after_first_call_in_seconds=100,
     )
     def _blob_rate_limit(self, func, *args, **kwargs):
         # `func_name` is a create/update/delete function.
@@ -209,7 +228,10 @@ class GcpBlobUpath(BlobUpath):
         # https://cloud.google.com/storage/docs/copying-renaming-moving-objects
         try:
             self.bucket.copy_blob(
-                self.blob(), target.bucket, target.blob_name, client=self.client,
+                self.blob(),
+                target.bucket,
+                target.blob_name,
+                client=self.client,
                 if_generation_match=None if overwrite else 0,
             )
         except NotFound:
@@ -218,14 +240,19 @@ class GcpBlobUpath(BlobUpath):
             raise FileExistsError(target)
 
     @overrides
-    def _export_file(self, target: Upath, *, overwrite=False) -> None:
+    def export_file(self, target: Upath, *, overwrite=False) -> None:
         if not isinstance(target, LocalUpath):
-            return super()._export_file(target, overwrite=overwrite)
+            return super().export_file(target, overwrite=overwrite)
+
+        # File download.
+
         if not overwrite and target.is_file():
             raise FileExistsError(target)
         os.makedirs(str(target.parent), exist_ok=True)
         try:
-            with open(target.localpath, 'wb') as file_obj:
+            with open(target.localpath, "wb") as file_obj:
+                # If `target` is an existing directory,
+                # will raise `IsADirectoryError`.
                 self._read_into_buffer(file_obj)
             updated = self.blob().updated
             if updated is not None:
@@ -253,15 +280,17 @@ class GcpBlobUpath(BlobUpath):
         # My experiments showed that `ctime` and `mtime` are equal.
 
     @overrides
-    def _import_file(self, source: Upath, *, overwrite=False) -> None:
+    def import_file(self, source: Upath, *, overwrite=False) -> None:
         if not isinstance(source, LocalUpath):
-            return super()._import_file(source, overwrite=overwrite)
+            return super().import_file(source, overwrite=overwrite)
+
+        # File upload.
 
         filename = str(source)
         content_type = self.blob()._get_content_type(None, filename=filename)
 
         def _upload():
-            with open(filename, 'rb') as file_obj:
+            with open(filename, "rb") as file_obj:
                 total_bytes = os.fstat(file_obj.fileno()).st_size
                 self._write_from_buffer(
                     file_obj,
@@ -278,6 +307,18 @@ class GcpBlobUpath(BlobUpath):
         # by other clients.
         return self.blob().exists(self.client)
 
+    @overrides
+    def is_dir(self) -> bool:
+        prefix = self.blob_name + "/"
+        blobs = self.client.list_blobs(
+            self.bucket,
+            prefix=prefix,
+            max_results=1,
+            page_size=1,
+            fields="items(name),nextPageToken",
+        )
+        return len(list(blobs)) > 0
+
     @overrides
     def iterdir(self):
         # From Google doc:
@@ -310,15 +351,15 @@ class GcpBlobUpath(BlobUpath):
         #
         # Search "List the objects in a bucket using a prefix filter | Cloud Storage"
 
-        prefix = self.blob_name + '/'
+        prefix = self.blob_name + "/"
         k = len(prefix)
-        blobs = self.client.list_blobs(self.bucket, prefix=prefix, delimiter='/')
+        blobs = self.client.list_blobs(self.bucket, prefix=prefix, delimiter="/")
         for p in blobs:
             obj = self / p.name[k:]  # files
             obj._blob = p
             yield obj
         for p in blobs.prefixes:
-            yield self / p[k:].rstrip('/')  # "subdirectories"
+            yield self / p[k:].rstrip("/")  # "subdirectories"
 
     def _acquire_lease(self, *, timeout: int = None):
         # Note: `timeout = None` does not mean infinite wait.
@@ -326,25 +367,29 @@ class GcpBlobUpath(BlobUpath):
         # just pass in a large number. Because user often associate
         # `timeout = None` with infinite wait, the default wait
         # is a long period.
-        if self._path == '/':
+        if self._path == "/":
             raise UnsupportedOperation("can not write to root as a blob", self)
         if timeout is None:
-            timeout = 3600  # seconds
+            timeout = 300  # seconds
 
         @opnieuw.retry(
-            retry_on_exceptions=(PreconditionFailed, FileExistsError),
+            retry_on_exceptions=(
+                *RETRY_WRITE_ON_EXCEPTIONS,
+                PreconditionFailed,
+                FileExistsError,
+            ),
             max_calls_total=10,
             retry_window_after_first_call_in_seconds=timeout,
         )
         def _acquire_():
-            self._blob_rate_limit(self._write_bytes, b'0')
+            self._write_bytes(b"0")
             self._generation = self.blob().generation
 
         t0 = time.perf_counter()
         try:
             _acquire_()
         except Exception as e:
-            raise LockAcquisitionTimeoutError(self, time.perf_counter() - t0) from e
+            raise LockAcquireError(self, time.perf_counter() - t0) from e
 
     @contextlib.contextmanager
     @overrides
@@ -374,13 +419,13 @@ class GcpBlobUpath(BlobUpath):
                         if_generation_match=self._generation,
                     )
                 except Exception as e:
-                    logger.error(e)
+                    raise LockReleaseError(f"failed to delete lock file {self}") from e
 
-    def open(self, mode='r', **kwargs):
-        '''
+    def open(self, mode="r", **kwargs):
+        """
         Use this on a blob (not a "directory") as a context manager.
         See Google documentation.
-        '''
+        """
         return self.blob().open(mode, **kwargs)
 
     def _read_into_buffer(self, file_obj):
@@ -388,7 +433,7 @@ class GcpBlobUpath(BlobUpath):
         if not file_info:
             raise FileNotFoundError(self)
         file_size = file_info.size  # bytes
-        if file_size <= MEGABYTES32:
+        if file_size <= LARGE_FILE_SIZE:
             try:
                 self.blob().download_to_file(file_obj, client=self.client)
                 return
@@ -412,17 +457,24 @@ class GcpBlobUpath(BlobUpath):
             k = 0
             p = 0
             while True:
-                kk = min(k + MEGABYTES32, file_size)
+                kk = min(k + LARGE_FILE_SIZE, file_size)
                 p += 1
-                yield (_download, (client, blob, k, kk - 1), {}, f"part {p}")
+                yield (
+                    _download,
+                    (client, blob, k, kk - 1),
+                    {},
+                    f"part {p}",
+                )
                 k = kk
                 if k >= file_size:
                     break
 
-        for buf, k in self._run_in_executor(_do_download(), f'downloading {self.name} - '):
+        for buf, k in self._run_in_executor(_do_download(), f"Downloading {self!r}"):
             n = file_obj.write(buf.getbuffer())
             if n != k:
-                raise BufferError(f"expecting to read {k} bytes; actually read {n} bytes")
+                raise BufferError(
+                    f"expecting to read {k} bytes; actually read {n} bytes"
+                )
             buf.close()
 
     @overrides
@@ -442,7 +494,7 @@ class GcpBlobUpath(BlobUpath):
 
     @overrides
     def riterdir(self):
-        prefix = self.blob_name + '/'
+        prefix = self.blob_name + "/"
         k = len(prefix)
         for p in self.client.list_blobs(self.bucket, prefix=prefix):
             obj = self / p.name[k:]
@@ -451,15 +503,20 @@ class GcpBlobUpath(BlobUpath):
 
     @overrides
     def with_path(self, *paths: str):
-        obj = self.__class__(*paths, bucket_name=self.bucket_name,
-                             project_id=self._project_id,
-                             credentials=self._credentials,)
-        obj._client = self._client
-        obj._bucket = self._bucket
+        obj = self.__class__(
+            *paths,
+            bucket_name=self.bucket_name,
+            project_id=self._project_id,
+            credentials=self._credentials,
+        )
+        obj._client = self.client
+        obj._bucket = self.bucket
         return obj
 
-    def _write_from_buffer(self, file_obj, *, overwrite=False, content_type=None, size=None):
-        if self._path == '/':
+    def _write_from_buffer(
+        self, file_obj, *, overwrite=False, content_type=None, size=None
+    ):
+        if self._path == "/":
             raise UnsupportedOperation("can not write to root as a blob", self)
 
         try:
@@ -480,11 +537,11 @@ class GcpBlobUpath(BlobUpath):
     def _write_bytes(self, data, **kwargs):
         b = BytesIO(data)
         b.seek(0)
-        self._write_from_buffer(b, content_type='text/plain', size=len(data), **kwargs)
+        self._write_from_buffer(b, content_type="text/plain", size=len(data), **kwargs)
 
     @overrides
     def write_bytes(self, data: Union[bytes, BufferedReader], *, overwrite=False):
         if isinstance(data, bytes):
             self._blob_rate_limit(self._write_bytes, data, overwrite=overwrite)
             return
-        self._write_from_buffer(data, content_type='text/plain', overwrite=overwrite)
+        self._write_from_buffer(data, content_type="text/plain", overwrite=overwrite)
